puts "A"
puts "B"

puts "C"


puts "D"



puts "E"




puts "F"

puts "This should be okay:



hello!"

proc okay_empty_proc {} {


}

proc not_okay_empty_proc {} {



}


# comments get counted as non-blank


if {[ensure_cmd_sub_ok]} {



    puts "too many lines at beginning and end of block"



}

switch $arg {


    "a" {
        puts "a"



    }



    "b" {
        puts "b"
    }



}

# too many lines at end of file



